--[[
Copyright (c) 2009, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

if( not CalendarBroker ) then
    return;
end--if

local L = LibStub( "AceLocale-3.0" ):GetLocale( "Calendar_Broker" );
local Z = LibStub( "LibBabble-Zone-3.0" ):GetLookupTable();

local TR = LibStub( "LibLordFarlander-TableRecycler-1.0", true );

if( not TR ) then
    return;
end--if

local WoWCalendarHandler = { ignoreEvent = 0, };
local CALENDAR_NAME = L["World of Warcraft"];

local ZONE_LOOKUP = {
-- Classic Dungeons
    ["BlackfathomDeeps"] = Z["Blackfathom Deeps"],
    ["BlackrockDepths"] = Z["Blackrock Depths"],
    ["BlackrockSpire"] = Z["Blackrock Spire"],
    ["Deadmines"] = Z["The Deadmines"],
    ["DireMaul"] = Z["Dire Maul"],
    ["Gnomeregan"] = Z["Gnomeregan"],
    ["Maraudon"] = Z["Maraudon"],
    ["RagefireChasm"] = Z["Ragefire Chasm"],
    ["RazorfenDowns"] = Z["Razorfen Downs"],
    ["RazorfenKraul"] = Z["Razorfen Kraul"],
    ["ScarletMonastery"] = Z["Scarlet Monastery"],
    ["Scholomance"] = Z["Scholomance"],
    ["ShadowFangKeep"] = Z["Shadowfang Keep"],
    ["StormwindStockades"] = Z["The Stockade"],
    ["Stratholme"] = Z["Stratholme"],
    ["SunkenTemple"] = Z["The Temple of Atal'Hakkar"],
    ["Uldaman"] = Z["Uldaman"],
    ["WailingCaverns"] = Z["Wailing Caverns"],
    ["ZulFarak"] = Z["Zul'Farrak"],
-- Classic Raids
    ["AQRuins"] = Z["Ahn'Qiraj"],
    ["BlackwingLair"] = Z["Blackwing Lair"],
    ["MoltenCore"] = Z["Molten Core"],
    ["OnyxiasLair"] = Z["Onyxia's Lair"],
    ["AQTemple"] = Z["Temple of Ahn'Qiraj"],
    ["ZulGurub"] = Z["Zul'Gurub"],
-- BC Dungeons
    ["Auchindoun"] = Z["Auchindoun"],
    ["CavernsOfTime"] = Z["Caverns of Time"],
    ["CoilFang"] = Z["Coilfang Reservoir"],
    ["HellfireCitadel"] = Z["Hellfire Citadel"],
-- BC Raids
    ["BlackTemple"] = Z["Black Temple"],
    ["TheEye"] = Z["The Eye"],
    ["GruulsLair"] = Z["Gruul's Lair"],
    ["HyjalPast"] = Z["Hyjal Summit"],
    ["Karazhan"] = Z["Karazhan"],
    ["HellfireCitadelRaid"] = Z["Magtheridon's Lair"],
    ["SerpentshrineCavern"] = Z["Serpentshrine Cavern"],
    ["Sunwell"] = Z["Sunwell Plateau"],
    ["TempestKeep"] = Z["Tempest Keep"],
    ["ZulAman"] = Z["Zul'Aman"],
-- WotLK Dungeons
    ["Ahnkahet"] = Z["Ahn'kahet: The Old Kingdom"],
    ["AzjolNerub"] = Z["Azjol-Nerub"],
    ["TheCullingofStratholme"] = Z["The Culling of Stratholme"],
    ["DrakTharon"] = Z["Drak'Tharon Keep"],
    ["Gundrak"] = Z["Gundrak"],
    ["HallsofLightning"] = Z["Halls of Lightning"],
    ["HallsofStone"] = Z["Halls of Stone"],
    ["TheNexus"] = Z["The Nexus"],
    ["TheOculus"] = Z["The Oculus"],
    ["Utgarde"] = Z["Utgarde Keep"],
    ["UtgardePinnacle"] = Z["Utgarde Pinnacle"],
    ["TheVioletHold"] = Z["The Violet Hold"],
    ["ArgentDungeon"] = Z["Trial of the Champion"],
-- WotLK Raids
    ["Malygos"] = Z["The Eye of Eternity"],
    ["IcecrownCitadel"] = Z["Icecrown Citadel"],
    ["Naxxramas"] = Z["Naxxramas"],
    ["ChamberOfAspects"] = Z["The Obsidian Sanctum"],
    ["Ulduar"] = Z["Ulduar"],
    ["VaultOfArchavon"] = Z["Vault of Archavon"],
    ["ArgentRaid"] = Z["Trial of the Crusader"],
};

local EVENT_TYPE_MAP = {
    [CALENDAR_EVENTTYPE_RAID or 1]    = nil,
    [CALENDAR_EVENTTYPE_DUNGEON or 2] = nil,
    [CALENDAR_EVENTTYPE_PVP or 3]     = L["PVP"],
    [CALENDAR_EVENTTYPE_MEETING or 4] = L["Meeting"],
    [CALENDAR_EVENTTYPE_OTHER or 5]   = nil,
};

local INVITE_TYPE_MAP = {
    [CALENDAR_INVITESTATUS_INVITED or 1]   = L["Invited"],
    [CALENDAR_INVITESTATUS_ACCEPTED or 2]  = L["Accepted"],
    [CALENDAR_INVITESTATUS_DECLINED or 3]  = L["Declined"],
    [CALENDAR_INVITESTATUS_CONFIRMED or 4] = L["Confirmed"],
    [CALENDAR_INVITESTATUS_OUT or 5]       = L["Out"],
    [CALENDAR_INVITESTATUS_STANDBY or 6]   = L["Stand by"],
};

function WoWCalendarHandler:PLAYER_ENTERING_WORLD( calendarMain )
    -- Do the initial scan
    self:FullScan( calendarMain );
    CalendarBroker:RegisterEvent( "CALENDAR_UPDATE_EVENT_LIST", self.CALENDAR_UPDATE_EVENT_LIST, calendarMain );
end--WoWCalendarHandler:PLAYER_ENTER_WORLD( calendarMain )

function WoWCalendarHandler.CALENDAR_UPDATE_EVENT_LIST( calendarMain )
    if( WoWCalendarHandler.ignoreEvent > 0 ) then
        WoWCalendarHandler.ignoreEvent = WoWCalendarHandler.ignoreEvent - 1;
    else
        calendarMain:CalendarUpdate( CALENDAR_NAME );
    end--if
end--WoWCalendarHandler.CALENDAR_UPDATE_EVENT_LIST()

function WoWCalendarHandler:GetWoWEventsForMonth( calendarMain, monthOffset, ourevents, counts, startDay )
    local month, year, days, _ = CalendarGetMonth( monthOffset );

    for day = startDay or 1, days do
        for i = 1, CalendarGetNumDayEvents( monthOffset, day ) do
            local title, hour, minute, calendarType, sequenceType, eventType, texture, modStatus, inviteStatus, invitedBy, difficulty, inviteType, sequenceIndex, numSequenceDays, difficultyName = CalendarGetDayEvent( monthOffset, day, i );
            local startTime = calendarMain.SecondsFromEpoch( month, day, year, hour, minute, 0 );
            local eventID = ("%i:%i:%i"):format( month, day, i );
            local instance = texture;

            --Figure out instance from the texture name
            if( instance and ( instance ~= "" ) ) then
                instance = (" (%s)"):format( ZONE_LOOKUP[instance] or instance );
            end--if
            if( ( calendarType == "RAID_RESET" ) or ( calendarType == "RAID_LOCKOUT" ) ) then
                title = ("%s %s"):format( GetDungeonNameWithDifficulty( title, difficultyName ), L["unlocks"] );
                calendarMain:AddEvent( startTime, 1 * CALENDARBROKER_SECPERHOUR, title, "RAID", CALENDAR_NAME, eventID );
            elseif( calendarType == "HOLIDAY" ) then
                local event = ourevents[title .. tostring( counts[title] or 0 )];
                local useTitle = title;
                local location = texture:match( "Calendar_DarkmoonFaire(%w+)" );

                if( location ) then
                    -- Darkmoon Faire, get the location of where it is for this time from the texture's name
                    if( location:find( "Elwynn" ) ) then
                        useTitle = ("%s (%s)"):format( title, Z["Elwynn Forest"] );
                    elseif( location:find( "Mulgore" ) ) then
                        useTitle = ("%s (%s)"):format( title, Z["Mulgore"] );
                    elseif( location:find( "Terokkar" ) ) then
                        useTitle = ("%s (%s)"):format( title, Z["Terokkar Forest"] );
                    end--if
                end--if                    

                if( sequenceType == "START" ) then
                    eventID = title;
                    ourevents[title .. tostring( counts[title] or 0 )] = TR:GetTable( "title", useTitle, "start", startTime, "length", ( 23 - hour ) * CALENDARBROKER_SECPERHOUR + ( 60 - minute ) * CALENDARBROKER_SECPERMIN, "type", calendarType, "id", eventID );
                elseif( sequenceType == "END" ) then               
                    if( event ) then
                        event.length = event.length + hour * CALENDARBROKER_SECPERHOUR + minute * CALENDARBROKER_SECPERMIN;
                        event.ended = true;
                        calendarMain:AddEvent( event.start, event.length, event.title, event.type, CALENDAR_NAME, event.id );
                        TR:ReturnTable( event );
                        ourevents[title .. tostring( counts[title] or 0 )] = nil;
                    else
                        eventID = title;
                        calendarMain:AddEvent( startTime, ( 23 - hour ) * CALENDARBROKER_SECPERHOUR + ( 60 - minute ) * CALENDARBROKER_SECPERMIN, useTitle, calendarType, CALENDAR_NAME, eventID );
                    end--if
                    counts[title] = ( counts[title] or 0 ) + 1;
                else
                    -- Not sure what to do about "INFO"
                    if( event ) then
                        event.length = event.length + 24 * CALENDARBROKER_SECPERHOUR;
                        if( not event.start ) then
                            event.start = startTime;
                        end--if
                    else
                        local length = ( ( hour == 14 ) and ( minute == 0 ) ) and ( 2 * CALENDARBROKER_SECPERHOUR ) or
                            ( ( 23 - hour ) * CALENDARBROKER_SECPERHOUR + ( 60 - minute ) * CALENDARBROKER_SECPERMIN );

                        if( sequenceType == "ONGOING" ) then                           
                            eventID = title;
                            ourevents[title .. tostring( counts[title] or 0 )] = TR:GetTable( "title", useTitle, "start", startTime, "length", length, "type", calendarType, "id", eventID );                            
                        else
                            calendarMain:AddEvent( startTime, length, useTitle, calendarType, CALENDAR_NAME, eventID );
                            counts[title] = ( counts[title] or 0 ) + 1;
                        end--if
                    end--if
                end--if
            elseif( ( calendarType == "PLAYER" ) and ( inviteStatus ~= CALENDAR_INVITESTATUS_DECLINED ) and ( inviteStatus ~= ( CALENDAR_INVITESTATUS_OUT or 5 ) ) ) then
                local eventTypeDesc = EVENT_TYPE_MAP[eventType] and (" (%s)"):format( EVENT_TYPE_MAP[eventType] ) or "";

                calendarMain:AddEvent( startTime, 1 * CALENDARBROKER_SECPERHOUR, title .. eventTypeDesc .. instance, calendarType, CALENDAR_NAME, eventID );
            elseif( ( calendarType == "GUILD" ) and ( inviteStatus ~= CALENDAR_INVITESTATUS_DECLINED ) and ( inviteStatus ~= ( CALENDAR_INVITESTATUS_OUT or 5 ) ) ) then
                local eventTypeDesc = EVENT_TYPE_MAP[eventType] and (" (%s)"):format( EVENT_TYPE_MAP[eventType] ) or "";

                calendarMain:AddEvent( startTime, 1 * CALENDARBROKER_SECPERHOUR, title .. eventTypeDesc .. instance, calendarType, CALENDAR_NAME, eventID );
            elseif( calendarType == "SYSTEM" ) then
                calendarMain:AddEvent( startTime, 1 * CALENDARBROKER_SECPERHOUR, title, calendarType, CALENDAR_NAME, eventID );
            elseif( calendarType == "ARENA" ) then
                calendarMain:AddEvent( startTime, 1 * CALENDARBROKER_SECPERHOUR, title, calendarType, CALENDAR_NAME, eventID );
            end--if
        end--for
    end--for
    return days - ( startDay or -1 ) + 1;
end--WoWCalendarHandler:GetWoWEventsForMonth( calendarMain, month, ourevents, counts )

function WoWCalendarHandler:FullScan( calendarMain )
    local ourevents = TR:GetTable();
    local counts = TR:GetTable();
    local _, month, day, year = CalendarGetDate();
    local prevmonth, prevyear, _, _ = CalendarGetMonth();
    local monthOffset = 0;
    local days = 0;
    local scanned = false;

    CalendarSetAbsMonth( month, year );
    self.ignoreEvent = self.ignoreEvent + 1;    
    while days < 120 do
        days = days + self:GetWoWEventsForMonth( calendarMain, monthOffset, ourevents, counts, ( monthOffset == 0 ) and day or 1 );
        monthOffset = monthOffset + 1;
        CalendarSetMonth( monthOffset );
        self.ignoreEvent = self.ignoreEvent + 1;
    end--while

    for _, event in pairs( ourevents ) do
        -- Find end of events that we didn't find the end of
        if( not scanned ) then
            monthOffset = monthOffset - 1;

            CalendarSetMonth( monthOffset );
            self.ignoreEvent = self.ignoreEvent + 1;            
            local month, year, days, firstday = CalendarGetMonth( monthOffset );

            for day = 1, days do
                for i = 1, CalendarGetNumDayEvents( monthOffset, day ) do
                    local title, hour, minute, calendarType, sequenceType, eventType, texture, modStatus, inviteStatus, invitedBy, difficulty = CalendarGetDayEvent( monthOffset, day, i );
                    
                    if( calendarType == "HOLIDAY" ) then
                        local event = ourevents[title .. tostring( counts[title] or 0 )];

                        if( event ) then
                            if( sequenceType == "END" ) then
                                if( event ) then
                                    event.length = event.length + hour * CALENDARBROKER_SECPERHOUR + minute * CALENDARBROKER_SECPERMIN;
                                    event.ended = true;
                                end--if
                                counts[title] = ( counts[title] or 0 ) + 1;
                            elseif( sequenceType ~= "START" ) then
                                event.length = event.length + 24 * CALENDARBROKER_SECPERHOUR;
                            end--if
                        end--if
                    end--if
                end--for
            end--for
            scanned = true;
        end--if
        calendarMain:AddEvent( event.start, event.length, event.title, event.type, CALENDAR_NAME, event.id );
    end--for
    TR:ReturnTable( ourevents );
    TR:ReturnTable( counts );
    CalendarSetAbsMonth( prevmonth, prevyear );
    self.ignoreEvent = self.ignoreEvent + 1;    
end--WoWCalendarHandler:FullScan( calendarMain )

function WoWCalendarHandler:GetCalendar( calendarMain, MaxNumberOfDays )
    local ourevents = TR:GetTable();
    local counts = TR:GetTable();
    local month, year, numdays, _ = CalendarGetMonth();
    local startTime = calendarMain.SecondsFromEpoch( month, 1, year, 0, 0, 0 );
    local endTime = calendarMain.SecondsFromEpoch( month, numdays, year, 23, 59, 59 );    
    local events = calendarMain.events;
    
    --Remove all events for the current month that do not spill over
    for index, event in ipairs( events ) do
        if( ( event.source == CALENDAR_NAME ) and  ( event.type ~= "HOLIDAY" ) and ( event.startTime <= startTime ) and ( event.endTime >= endTime ) ) then
            table.remove( events, index );
        end--if
    end--for
    self:GetWoWEventsForMonth( calendarMain, 0, ourevents, counts, 1 );
    TR:ReturnTable( ourevents );
    TR:ReturnTable( counts );    
end--WoWCalendarHandler:GetCalendar( calendarMain, MaxNumberOfDays )

function WoWCalendarHandler:GetEventInfo( calendarMain, id )
    -- Unused ATM and needs to be written correctly, this code is NOT correct
    local _, month, day, year = CalendarGetDate();
    local prevmonth, prevyear, _, _ = CalendarGetMonth();
    local returnVal = TR:GetTable();
    local eventMonth, eventDay, eventIndex = id:match( "(%d+):(%d+):(%d+)" );
    
    CalendarSetAbsMonth( month, year );
    CalendarOpenEvent( eventMonth, eventDay, eventIndex );
    
    local title, description, creator, eventType, repeatOption, maxSize, textureIndex,
        weekday, month, day, year, hour, minute,
        lockoutWeekday, lockoutMonth, lockoutDay, lockoutYear, lockoutHour, lockoutMinute,
        locked, autoApprove, pendingInvite, inviteStatus = CalendarGetEventInfo();
        
    returnVal.description = title;
    returnVal.invited = pendingInvite;
    returnVal.inviteStatus = inviteStatus;

    CalendarCloseEvent();
    CalendarSetAbsMonth( prevmonth, prevyear );
    return returnVal;
end--WoWCalendarHandler:GetEventInfo( calendarMain, id )

CalendarBroker.Calendars[CALENDAR_NAME] = WoWCalendarHandler;
